/*
    futebus - FUse TExt BUS - fuse based IPC bus for text lines
    Copyright (C) 2023  Tibor 'Igor2' Palinkas

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along
    with this program; if not, write to the Free Software Foundation, Inc.,
    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

    Project page: http://www.repo.hu/projects/futebus
    VCS: svn://svn.repo.hu/futebus/trunk
    Contact: see doc/contact.html
*/

#define _FILE_OFFSET_BITS 64
#define FUSE_USE_VERSION 28
#include <string.h>
#include <fuse.h>
#include <poll.h>
#include <errno.h>
#include "genvector/gds_char.h"

#include "futebus.h"
#include "dispatch.h"
#include "channel.h"
#include "stat.h"

char *vers = "1.0.0";

FILE *logfile;
size_t msg_size_max = 1024;
size_t write_size_max = 65536;
int foreground;
long rlist_size_max = 128;

gdl_list_t channels;
gdl_list_t zombie_channels;
mutex_t channels_lock;

static int futebus_readdir(const char *path, void *buf, fuse_fill_dir_t filler, off_t offset, struct fuse_file_info *fi)
{
	channel_t *c;
	logprintf("readdir '%s'\n", path);

	filler(buf, ".", NULL, 0);
	filler(buf, "..", NULL, 0);
	filler(buf, "_stat", NULL, 0);

	for(c = gdl_first(&channels); c != NULL; c = gdl_next(&channels, c))
		filler(buf, c->path, NULL, 0);


	return 0;
}

static int futebus_open(const char *path, struct fuse_file_info *fi)
{
	uint64_t fh;
	int rd = 0, wr = 0, as_master = 0, is_stat = 0, err;

	if (*path == '/') path++;
	if (*path == '.') { path++; as_master = 1; }
	else if (strcmp(path, "_stat") == 0) is_stat = 1;

	switch(fi->flags & O_ACCMODE) {
		case O_RDONLY: rd = 1; logprintf("open %s RO\n", path);  break;
		case O_WRONLY: wr = 1; logprintf("open %s WO\n", path); break;
		case O_RDWR:   rd = wr = 1; logprintf("open %s rw\n", path);  break;
	}

	if (is_stat) {
		if (!rd || wr)
			return -ESRCH;
		if (stat_open(path, &fh) != 0)
			return -EIO;
	}
	else if ((err = chan_open(path, &fh, rd, wr, as_master)) != 0)
		return err;

	logprintf("succesful open, fh=%llx as %s\n", (unsigned long long)fh, is_stat ? "stat" : "channel");
	fi->fh = fh;
	return 0;
}

static int futebus_release(const char *path, struct fuse_file_info *fi)
{
	uint32_t chanID, clientID;

	fd_to_chan_client(fi->fh, chanID, clientID);

	logprintf("Release %llx\n", (unsigned long long)fi->fh);

	return dispatch_client_release(chanID, clientID);
}

static int futebus_read(const char *path, char *buf, size_t size, off_t offs, struct fuse_file_info *fi)
{
	uint32_t chanID, clientID;

	fd_to_chan_client(fi->fh, chanID, clientID);

	logprintf("Read %llx: size=%ld offs=%ld\n", (unsigned long long)fi->fh, (long)size, (long)offs);
	return dispatch_client_read(chanID, clientID, buf, size, fi->flags & O_NONBLOCK);
}

static int futebus_write(const char *path, const char *buf, size_t size, off_t offs, struct fuse_file_info *fi)
{
	uint32_t chanID, clientID;

	fd_to_chan_client(fi->fh, chanID, clientID);

	if (logfile != NULL) {
		char tmp[1024];
		int len = size < sizeof(tmp)-1 ? size : sizeof(tmp)-1;
		strncpy(tmp, buf, len);
		tmp[len] = '\0';
		logprintf("Write %llx #%ld:%ld: size=%ld offs=%ld: %s\n", (unsigned long long)fi->fh, (long)chanID, (long)clientID, (long)size, (long)offs, tmp);
	}
	return dispatch_client_write(chanID, clientID, buf, size, fi->flags & O_NONBLOCK);
}

static int futebus_getattr(const char *path, struct stat *stbuf)
{
	logprintf("getattr '%s'\n", path);

	memset(stbuf, 0, sizeof(struct stat));

	if (strcmp(path, "/") == 0) {
		stbuf->st_mode = S_IFDIR | 0777;
		stbuf->st_nlink = 2;
	}
	else {
		stbuf->st_mode = S_IFREG | 0644;
		stbuf->st_nlink = 1;
		stbuf->st_size = 0;
		stbuf->st_uid = 1000;
		stbuf->st_gid = 1000;
		stbuf->st_mtime = 0;
		stbuf->st_ctime = 0;
	}

	return 0;
}

void futebus_poll_uninit(client_t *client)
{
	mutex_lock(&client->poll_lock);
	fuse_pollhandle_destroy(client->poll_handle);
	client->poll_handle = NULL;
	client->poll_revents = NULL;
	mutex_unlock(&client->poll_lock);
}

static void poll_notify(client_t *client)
{
	fuse_notify_poll(client->poll_handle);
	fuse_pollhandle_destroy(client->poll_handle);
	client->poll_revents = NULL;
	client->poll_handle = NULL;
}

void futebus_poll_readable(client_t *client)
{
	mutex_lock(&client->poll_lock);
	if (client->poll_handle != NULL) {
		logprintf("poll notify: #%ld POLLIN\n", (long)client->id);
		(*client->poll_revents) |= POLLIN;
		poll_notify(client);
	}
	mutex_unlock(&client->poll_lock);
}



int futebus_poll(const char *path, struct fuse_file_info *fi, struct fuse_pollhandle *ph, unsigned *reventsp)
{
	uint32_t chanID, clientID;
	channel_t *chan;
	client_t *client;

	fd_to_chan_client(fi->fh, chanID, clientID);
	chan = id2chan(chanID);
	client = id2client(chan, clientID);
	if (client == NULL)
		return -ENODEV;

	mutex_lock(&client->poll_lock);
	if (ph != NULL) {
		if (client->poll_handle != NULL)
			fuse_pollhandle_destroy(client->poll_handle);
		client->poll_handle = ph;
		client->poll_revents = reventsp;
	}
	mutex_unlock(&client->poll_lock);

	mutex_lock(&client->rlist_lock);
	if ((client->rlist.used > 0) || (client->rbuf.used > 0)) {
		(*reventsp) |= POLLIN | POLLOUT;
		logprintf("poll return: #%ld POLLIN\n", (long)client->id);
	}
	mutex_unlock(&client->rlist_lock);

	return 0;
}

static void futebus_destroy(void *private_data)
{
	logprintf("Destroy all.\n");
	stat_uninit();
	dispatch_uninit();
}


static struct fuse_operations futebus_oper = {
	.readdir  = futebus_readdir,
	.open     = futebus_open,
	.release  = futebus_release,
	.read     = futebus_read,
	.write    = futebus_write,
	.getattr  = futebus_getattr,
	.poll     = futebus_poll,
	.destroy  = futebus_destroy
};

/* convert decimal numeric to long with minimum value; return 1 on success, -1
   on error. */
static int cli_parse_long(long *dst, const char *key, const char *src, size_t min_val)
{
	char *end;
	long l = strtol(src, &end, 10);
	if (*end != '\0') {
		fprintf(stderr, "Invalid value for %s: expected a decimal integer, got '%s'\n", key, src);
		return -1;
	}
	if (l < min_val) {
		fprintf(stderr, "Invalid value for %s: needs to be at least %ld\n", key, (long)min_val);
		return -1;
	}
	*dst = l;
	return 1;
}

/* Parse a single key=value option; return 1 if handled, 0 if not handled
   but do not throw error on unknonw keys (they are for fuse). Return -1
   when option is recongized but value is invalid */
static int cli_parse_opt(char *key, char *val)
{
	int res;
	long l;

	if (strcmp(key, "msg_size_max") == 0) {
		res = cli_parse_long(&l, key, val, 4);
		msg_size_max = l;
		return res;
	}

	if (strcmp(key, "write_size_max") == 0) {
		res = cli_parse_long(&l, key, val, 4);
		write_size_max = l;
		return res;
	}

	if (strcmp(key, "rlist_size_max") == 0) {
		res = cli_parse_long(&l, key, val, 2);
		rlist_size_max = l;
		return res;
	}

	return 0;
}


/* Parse -o payload, execute and remove futebus options, copy back fuse options
   to opts. Returns 1 if anything is left in opts */
static int cli_parse_opts(char *opts)
{
	gds_t tmp = {0};
	char *s, *next, *sep;

	for(s = opts; s != NULL; s = next) {
		int parsed = 0;

		next = strchr(s, ',');
		if (next != NULL) {
			*next = '\0';
			next++;
		}

		/* figure if it's something we canhandle */
		sep = strchr(s, '=');
		if (sep != NULL) {
			*sep = '\0';
			parsed = cli_parse_opt(s, sep+1);
			if (parsed < 0)
				exit(1);
			*sep = '=';
		}

		if (!parsed) { /* not for us, keep it */
			if (tmp.used > 0)
				gds_append(&tmp, ',');
			gds_append_str(&tmp, s);
		}
	}

	if (tmp.used != 0) {
		/* opts is large enough: we never add anything, only use the characters from opts */
		memcpy(opts, tmp.array, tmp.used+1);
		gds_uninit(&tmp);
		return 1;
	}
	return 0;
}

static void print_futebus_help(void)
{
	fprintf(stderr, "\n");
	fprintf(stderr, "futebus-specific command line arguments:\n");
	fprintf(stderr, "    -L logfile             log events to logfile\n");
	fprintf(stderr, "\n");
	fprintf(stderr, "futebus-specific -o options:\n");
	fprintf(stderr, "    -o msg_size_max=N      maximum size of msg, including the newline\n");
	fprintf(stderr, "    -o rlist_size_max=N    max number of msgs put on an rlist\n");
	fprintf(stderr, "    -o write_size_max=N    maximum size of a write()\n");
	fprintf(stderr, "\n");
}


#define ARGLEN 128
int main(int argc, char *argv[])
{
	char *args[ARGLEN], *mnt = NULL;
	int n, d, res, help = 0;


	if (argv[1] == NULL) {
		fprintf(stderr, "need a mount point\n");
		return 1;
	}

	args[0] = "fusebus";
	args[1] = "-o";
	args[2] = "direct_io"; /* allow short read and short write */
	for(d = 3, n = 1; n < argc; n++) {
		int copy = 1;
		char *cmd = argv[n];

		if (*cmd == '-') {
			while(*cmd == '-') cmd++;
			switch(*cmd) {
				case 'h':
					help = 1;
					break;
				case 'f':
					foreground = 1;
					logfile = stdout;
					break;
				case 'L':
					if (argv[n+1] == NULL) {
						fprintf(stderr, "-L needs a log file name\n");
						exit(1);
					}
					logfile = fopen(argv[n+1], "w");
					if (logfile == NULL) {
						fprintf(stderr, "-L failed to open log file '%s' for write\n", argv[n+1]);
						exit(1);
					}
					n++;
					copy = 0;
					break;
				case 'o':
					if (cli_parse_opts(argv[n+1])) {
						if (d == ARGLEN-2) goto too_many_args;
						args[d++] = "-o";
						args[d++] = argv[n+1];
					}
					n++;
					copy = 0;
					break;
			}
		}
		else {
			if (mnt != NULL) {
				fprintf(stderr, "Multiple mount points not supported (at '%s')\n", argv[n]);
				return 1;
			}
			else
				mnt = argv[n];
			copy = 0;
		}

		if (copy) {
			if (d == ARGLEN-1) goto too_many_args;
			args[d++] = argv[n];
		}
	}
	if (mnt != NULL)
		args[d++] = mnt;
	args[d] = NULL;

	mutex_init(&channels_lock);

	res = fuse_main(d, args, &futebus_oper, NULL);
	if (help)
		print_futebus_help();
	return res;

	too_many_args:;
	fprintf(stderr, "Too many command line arguments\n");
	return 1;
}

